/*  DiscreteCostFunction.h

    Emma Robinson, FMRIB Image Analysis Group

    Copyright (C) 2012 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk

    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK


    LICENCE

    FMRIB Software Library, Release 6.0 (c) 2018, The University of
    Oxford (the "Software")

    The Software remains the property of the Oxford University Innovation
    ("the University").

    The Software is distributed "AS IS" under this Licence solely for
    non-commercial use in the hope that it will be useful, but in order
    that the University as a charitable foundation protects its assets for
    the benefit of its educational and research purposes, the University
    makes clear that no condition is made or to be implied, nor is any
    warranty given or to be implied, as to the accuracy of the Software,
    or that it will be suitable for any particular purpose or for use
    under any specific conditions. Furthermore, the University disclaims
    all responsibility for the use which is made of the Software. It
    further disclaims any liability for the outcomes arising from using
    the Software.

    The Licensee agrees to indemnify the University and hold the
    University harmless from and against any and all claims, damages and
    liabilities asserted by third parties (including claims for
    negligence) which arise directly or indirectly from the use of the
    Software or the sale of any products based on the Software.

    No part of the Software may be reproduced, modified, transmitted or
    transferred in any form or by any means, electronic or mechanical,
    without the express permission of the University. The permission of
    the University is not required if the said reproduction, modification,
    transmission or transference is done without financial return, the
    conditions of this Licence are imposed upon the receiver of the
    product, and all original and amended source code is included in any
    transmitted product. You may be held legally responsible for any
    copyright infringement that is caused or encouraged by your failure to
    abide by these terms and conditions.

    You are not permitted under this Licence to use this Software
    commercially. Use for which any financial return is received shall be
    defined as commercial use, and includes (1) integration of all or part
    of the source code or the Software into a product for sale or license
    by or on behalf of Licensee to third parties or (2) use of the
    Software or any derivative of it for research with the final aim of
    developing software products for sale or license to a third party or
    (3) use of the Software or any derivative of it for research with the
    final aim of developing non-software products for sale or license to a
    third party, or (4) use of the Software to provide any service to an
    external organisation for which payment is received. If you are
    interested in using the Software commercially, please contact Oxford
    University Innovation ("OUI"), the technology transfer company of the
    University, to negotiate a licence. Contact details are:
    fsl@innovation.ox.ac.uk quoting Reference Project 9564, FSL.*/
#ifndef __DISCRETECOSTFUNCTION_H__
#define __DISCRETECOSTFUNCTION_H__

#include <vector>
#include "similarities.h"

#include <boost/variant/variant.hpp>
#include <boost/variant/get.hpp>

typedef std::map<std::string, boost::variant<int, std::string, double, float, bool> > myparam;
typedef std::pair<std::string, boost::variant<int, std::string, double, float, bool> > parameterPair;


namespace DISCRETEOPT{

  class DiscreteCostFunction
  {
  public:
    //typedef std::vector< std::vector<float> >		PairwiseMemory;
    //typedef std::vector< std::vector<float> >		TripletMemory;
    /**
     * Constructor.
     */
    DiscreteCostFunction();


    /**
     * Destructor.
     */
    virtual ~DiscreteCostFunction();

    /**
     * Resets all costs to zero.
     */
    void reset();

    /**
     * Sets the pairs index buffer.
     */
    void setPairs(int* p) { _pairs = p; }

    /**
     * Sets the pairs index buffer.
     */
    void setTriplets(int* p) { _triplets = p; }

    /**
     * Sets the quartets index buffer.
     */
    void setQuartets(int* p) { _quartets = p; }

    /**
     * Returns the unary costs look-up table.
     */
    double* getUnaryCosts() { return unarycosts; }

    /**
     * Returns the pairwise costs look-up table.
     */
    double* getPairwiseCosts() { return paircosts; }

    /**
     * Returns the triplet costs look-up table.
     */
    //double* getTripletCosts() { return tripletcosts; }

    /**
     * Returns the triplet costs look-up table.
     */
    // double* getQuartetCosts() { return quartetcosts; }

    /**
     * Computes the unary costs look-up table.
     */
    virtual void computeUnaryCosts(){};

    /**
     * Computes the unary potential for a the given node
     */
    virtual double computeUnaryCost(int node, int label){ return 0; };

    /**
     * Computes the pairwise costs look-up table.
     */
    virtual void computePairwiseCosts(const int *pairs){};

    /**
     * Computes the pairwise potential for a the given pair and labels.
     */
    virtual double computePairwiseCost(int pair, int labelA, int labelB){ return 0; };

    /**
     * Initializes the pair weights with given weights. If no weights are provided, all are set to one.
     */
    virtual void initPairWeights(const double *weights = 0);

    /**
     * Computes the triplet costs look-up table.
     */
    // virtual void computeTripletCosts() {};

    /**
     * Computes the triplet potential for a the given triplet and labels.
     */
    virtual double computeTripletCost(int triplet, int labelA, int labelB, int labelC) { return 0; }

    /**
     * Initializes the triplet weights with given weights. If no weights are provided, all are set to one.
     */
    ///virtual void initTripletWeights(const double *weights = 0);


    /**
     * Computes the quartet costs look-up table.
     */
    // virtual void computeQuartetCosts() {};

    /**
     * Computes the quartet potential for a the given triplet and labels.
     */
    virtual double computeQuartetCost(int quartet, int labelA, int labelB, int labelC, int labelD) { return 0; }


    /**
     * Evaluates the total cost for the zero labeling.
     */
    double evaluateTotalCostSumZeroLabeling();

    /**
     * Evaluates the total cost for the given labeling.
     */
    virtual double evaluateTotalCostSum(const int *labeling, const int *pairs, const int *triplets = 0,const int *quartets = 0);

    /**
     * Evaluates the sum of unary costs for a given labeling.
     */
    double evaluateUnaryCostSum(const int *labeling);

    /**
     * Evaluates the sum of pairwise costs for a given labeling.
     */
    double evaluatePairwiseCostSum(const int *labeling, const int *pairs);

    /**
     * Evaluates the sum of triplet costs for a given labeling.
     */
    double evaluateTripletCostSum(const int *labeling, const int *triplets);

    /**
     * Evaluates the sum of quartet costs for a given labeling.
     */
    double evaluateQuartetCostSum(const int *labeling, const int *quartets);

    /**
     * Enables the memory for the pairwise potentials computation.
     */
    // void enablePairwiseMemory();

    /**
     * Enables the memory for the triplet potentials computatio
     */
    // void enableTripletMemory();

    virtual void set_parameters(myparam & )=0;

    virtual void report(){};

  protected:
    /**
     * Initializes the cost function.
     * \param numNodes Number of nodes.
     * \param numLabels Number of labels.
     * \param numPairs Number of pairs.
     * \param numTriplets Number of triplets.
     */
    void initialize(int numNodes, int numLabels, int numPairs, int numTriplets = 0,int numQuartets = 0);

    int				m_num_nodes;			///< Number of nodes.
    int				m_num_labels;			///< Number of labels.
    int				m_num_pairs;			///< Number of pairs.
    int				m_num_triplets;			///< Number of triplets.
    int				m_num_quartets;			///< Number of quartets.


    double			*unarycosts;			///< Unary potentials look-up table.
    double			*paircosts;				///< Pairwise potentials look-up table.
    //  double			*tripletcosts;			///< Triplet potentials look-up table.
    //  double			*quartetcosts;			///< Quartet potentials look-up table.

    ///(NOT USED!!)
    double 			*pairweights;			///< Weightings for the pairwise potentials.
    //double 			*tripletweights;		///< Weightings for the triplet potentials.
    //

    int *_pairs;
    int *_triplets;
    int *_quartets;
    int *labels;							///< Labeling array.

    float _reglambda;  // scaling parameter for regulariser

    bool _debug;
    bool  _verbosity;
    bool _concat;
    std::string _outdir;
    std::string _matlabpath;
//bool write;
    //PairwiseMemory	m_paircosts_memory;		///< Pairwise potentials memory to avoid re-computations.
    //TripletMemory	m_tripletcosts_memory;	///< Triplet potentials memory to avoid re-computations.
    //bool			m_is_memory;			///< Indicates whether the pairwise potentials memory is enabled.
    //bool			m_is_memory_triplets;	///< Indicates whether the triplet potentials memory is enabled.
  };

  /// should be able to implement affine spherical registration using discrete labels represent fixed rotations
  class DummyCostFunction: public DiscreteCostFunction
  {
  public:
    DummyCostFunction(){m_num_labels=2;};

    void setUnaryCost(int node, double cost0, double cost1){

      unaryenergies.insert(std::pair<int, std::vector<double> >(node, std::vector<double>()));
      unaryenergies[node].push_back(cost0);
      unaryenergies[node].push_back(cost1);
    }

    void setPairwiseCost(int ind, double E00, double E01, double E10, double E11){
      pairenergies.insert(std::pair<int, std::vector<double> >(ind, std::vector<double>()));
      pairenergies[ind].push_back(E00);
      pairenergies[ind].push_back(E01);
      pairenergies[ind].push_back(E01);
      pairenergies[ind].push_back(E11);
    }

    double computePairwiseCost(int pair, int labelA, int labelB){
      double cost;
      if(labelA==0 && labelB==0)
	cost=pairenergies[pair][0];
      else if (labelA==0 && labelB==1)
	cost=pairenergies[pair][1];
      else if  (labelA==1 && labelB==0)
	cost=pairenergies[pair][2];
      else
	cost=pairenergies[pair][3];
      // cout << pair << " " << labelA << " " << labelB << " computePairwiseCost " << cost << endl;
      return cost;
    }

    void convertenergies(int numNodes,int numPairs, int numLabels){
      m_num_nodes=numNodes; m_num_labels=numLabels; m_num_pairs=numPairs;
      if (unarycosts){ delete [] unarycosts; }
      unarycosts = new double[numNodes*numLabels];
      for(int i=0;i<numLabels;i++){
	for(int j=0;j<numNodes;j++){
	  unarycosts[i*numNodes+j]=unaryenergies[j][i];
	}
      }
    }

    void reset(){unaryenergies.clear();  pairenergies.clear();
      // if (unarycosts){ delete [] unarycosts; }
    }
    void set_parameters(myparam & ){};
  protected:
    std::map<int,std::vector<double> > unaryenergies; /// maps of nodes  xlabels x vals
    std::map<int,std::vector<double> > pairenergies;
  };

  class SRegDiscreteCostFunction: public DiscreteCostFunction
  {

  public:
    //
    // Constructor.
    //
    SRegDiscreteCostFunction();

    ///////////////////// VARIABLE ASSIGNMENT //////////////////////////
    // data input and reference mesh, plus low resolution control point grid
    virtual void set_meshes(const NEWMESH::newmesh & target,const NEWMESH::newmesh & source, const NEWMESH::newmesh & GRID, int num=1){_TARGET=target; _SOURCE=source; _ORIG=source;  _CPgrid=GRID; _oCPgrid=GRID;}
    void set_meshes(const NEWMESH::newmesh & target,const NEWMESH::newmesh & source){_TARGET=target; _SOURCE=source;};

    void set_anatomical(const NEWMESH::newmesh &targetS,const NEWMESH::newmesh &targetA,const NEWMESH::newmesh & sourceS,const NEWMESH::newmesh & sourceA){_TARGEThi=targetS; _aTARGET=targetA;_aICO=sourceS, _aSOURCE=sourceA;}
    void set_anatomical_neighbourhood(const std::vector<std::map<int,double> > & weights, const std::vector<std::vector<int> > neighbourhood){_ANATbaryweights=weights; NEARESTFACES=neighbourhood; }

    ///// holds data for source and target mesh, used for look up during similarity estimation

    NEWMESH::newmesh project_anatomical();
    void set_featurespace(const std::shared_ptr<NEWMESH::featurespace> &features, bool _concatenate=false){FEAT=features; _concat=_concatenate;}

    void initialize_regulariser();

    void set_initial_angles(const std::vector<std::vector<double> >  angles){

      double meanang=0;
      for (size_t i = 0; i < angles.size(); i++)
      {
    	for (size_t j = 0; j < angles[i].size(); j++)
    	{
	      meanang+=angles[i][j];
    	}
      }
      _MEANANGLE=meanang/(3.0*angles.size());
    }

    virtual void set_matlab_path(std::string s){_matlabpath=s;}
    ///
    void set_dataaffintyweighting(const NEWMAT::Matrix &HRWeight){_HIGHREScfweight=HRWeight;}

    //// holds neighbourhood information
    virtual void set_relations(const std::shared_ptr<NEWMESH::RELATIONS> &CONTROL,const std::shared_ptr<NEWMESH::RELATIONS> &TARG){ _controlrel=CONTROL; _targetrel=TARG;
 _sourcerel=_controlrel->invert_relations(_CPgrid,_SOURCE);  }

    void set_parameters(myparam & );
    //MAXSEP is control grid vertex spacings
    virtual void set_spacings(const NEWMAT::ColumnVector &spacings, const double MAX, int num=0){MAXSEP=spacings;MVDmax=MAX;}

    void debug(){_debug=true;} // for debuging
    /// label list changes between iterations for spherical optimisation (barycentres and vertices of regular sampling grid)
    void set_labels(const std::vector<NEWMESH::Pt> labellist, const std::vector<NEWMAT::Matrix> ROT=std::vector<NEWMAT::Matrix>()){_labels=labellist;  ROTATIONS= std::shared_ptr<std::vector<NEWMAT::Matrix> >(new std::vector<NEWMAT::Matrix> (ROT));  } // ROTATION MATRIX MAY ONLY BE REQUIRED FOR NON LINEAR FRAMEWORK

    inline void set_iter(int iter) {_iter=iter;}
    virtual void reset_source(const NEWMESH::newmesh & source, int num=0){_SOURCE=source; }

    void reset_anatomical(const std::string &, const int &);

    //void set_pairtrIDs(const vector<int> &pair_trIDs){ _trIDs=pair_trIDs;}
    virtual void reset_CPgrid(const NEWMESH::newmesh & grid, int num=0){_CPgrid=grid;}

    //////////////// GENERIC HELPER FUNCTIONS

    bool within_controlpt_range(const int &CPindex, const int &sourceindex){
      double dist=0.0;
      if(within_controlpt_range(CPindex,sourceindex,dist))return true;
      else return false;

    }

    bool within_controlpt_range(const int &CPindex, const int &sourceindex, double & dist){
      NEWMESH::Pt CP=_CPgrid.get_coord(CPindex);
      dist= 2*RAD*asin((CP-_SOURCE.get_coord(sourceindex)).norm()/(2*RAD));
      if(dist<_controlptrange*MAXSEP(CPindex+1))return true;
      else return false;

    }
    virtual void initialize(int numNodes,int numLabels, int numPairs, int numTriplets = 0, int numQuartets = 0);

    virtual void resample_target_data(const NEWMESH::Pt &, const NEWMESH::Pt &, const NEWMESH::Pt &, const NEWMESH::Pt &, const int &, const int &, const int &, int &){};

    virtual void resample_weights(){};
    virtual void get_source_data(){};
    virtual void reset_target_data(int){};
    virtual double triplet_likelihood(const int &,const int &,const int &,const int &,const NEWMESH::Pt &,const NEWMESH::Pt &,const NEWMESH::Pt &){ return 0; };

    NEWMESH::newmesh get_SOURCE(){return _SOURCE;}
    void report(){ if(_debug) std::cout << " sumlikelihood " << sumlikelihood << " sumregcost " << sumregcost <<std::endl; }
  protected:

    //////////////////////// MESHES //////////////////////
    NEWMESH::newmesh _TARGET; // TARGET MESH
    NEWMESH::newmesh _TARGEThi; // TARGET MESH
    NEWMESH::newmesh _SOURCE; // SOURCE MESH
    NEWMESH::newmesh _aTARGET; // ANATOMICAL TARGET MESH

    NEWMESH::newmesh _aSOURCE; // ANATOMICAL SOURCE MESH
    NEWMESH::newmesh _aICO; // ANATOMICAL SOURCE MESH

    NEWMESH::newmesh _aSOURCEtrans; // ANATOMICAL SOURCE MESH


    NEWMESH::newmesh _ORIG; // NON DEFORMED SOURCE
    NEWMESH::newmesh _oCPgrid; ///// NON DEFORMED CP GRID
    NEWMESH::newmesh _CPgrid; ///// CONTRO

    //////////////// CF WEIGHTINGS ////////////////////////////
    NEWMAT::Matrix _HIGHREScfweight; /// source mesh cfweight

    NEWMAT::ColumnVector MAXSEP;

    /////////////////////// NEIGHBOURHOOD INFO ////////////////////
    std::shared_ptr<NEWMESH::RELATIONS> _controlrel; // hold control grid neighbours of each source vertex
    std::shared_ptr<NEWMESH::RELATIONS> _targetrel; // hold target grid neighbours of each source vertex
    NEWMESH::RELATIONS _sourcerel; // hold source grid neighbours of each CP(!!) vertex/triangle

    NEWMESH::RELATIONS _anatrel;

    std::vector<std::vector<int> > _sourceinrange;
    std::vector<std::vector<int> > NEARESTFACES;
    //////////////////////// FEATURE SET //////////////////
    std::shared_ptr<NEWMESH::featurespace> FEAT; /// holds data
    sparsesimkernel<double> sim; // similarity object

    /////////////////// LABELLING PARAMETERS////////
    std::vector<NEWMESH::Pt> _labels;
    std::shared_ptr<std::vector<NEWMAT::Matrix> > ROTATIONS; // rotates label set onto each control point

    double MVDmax; // max distance between CPs
    double anatMVD;
    double _MEANANGLE;
    float _controlptrange;
    double sumlikelihood;
    double sumregcost;

    float _mu; // shear modulus
    float _kappa; // bulk modulus
    float _pottsthreshold;
    float _rexp;
    float _sigma;

    ///////// USER DEFINED PARAMETERS /////////
    int _simmeasure;
    int _RES;
    int _aRES;
    int  _rmode;
    int _iter;
    int _threads;

    float _k_exp;
    std::vector<std::vector<double> > _sourcedata;
    std::vector<std::vector<double> > _targetdata;
    std::vector<std::vector<double> > _weights;

    std::vector<std::map<int,double> >  _ANATbaryweights;
    NEWMESH::resampler R;
    double MAXstrain;
    double strain95;
  };


  /*
      class AffineSRegDiscreteCostFunction: public SRegDiscreteCostFunction
      {
      public:
      //
      // Constructor.
      //
      AffineDiscreteCostFunction();

      //
      // Computes the unary costs look-up table.

      void computeUnaryCosts() {};

      //
      // Computes the pairwise costs look-up table.

      void computePairwiseCosts() {};

      protected:

      };
    */
  class NonLinearSRegDiscreteCostFunction: public SRegDiscreteCostFunction
  {

  protected:

    ////////////// REGULARISER OPTIONS /////////////////////////
    float _maxdist;
    float _expscaling;
    bool _dweight;
    bool _anorm;

    int _kNN;
    int _currentlabelA;
    int _currentlabelB;
    int _currentlabelC;

    NEWMAT::ColumnVector AbsoluteWeights;

    std::vector<NEWMAT::Matrix> PreviousDeformations;
    std::vector<NEWMAT::Matrix> ROTATE2LABEL;
    std::vector<NEWMESH::Pt> _ORIGpositions;

  public:
    //
    // Constructor.
    //
    NonLinearSRegDiscreteCostFunction();

    virtual void initialize(int numNodes, int numLabels, int numPairs, int numTriplets = 0,int numQuartets = 0); // quartets not used yet so no code for them below

    void set_parameters(myparam & );

    //// compute unary costs
    void computeUnaryCosts();

    //// compute pairwise costs
    double computePairwiseCost(int pair, int labelA, int labelB);
    void computePairwiseCosts(const int *pairs) ;

    //compute triplet costs
    // void computeTripletCosts();
    double computeTripletCost(int triplet, int labelA, int labelB, int labelC);
    virtual double triplet_likelihood(const int &,const int &,const int &,const int &,const NEWMESH::Pt &,const NEWMESH::Pt &,const NEWMESH::Pt &){ return 0;};
   /**
     * Computes the quartet costs look-up table.
     */
   // void computeQuartetCosts() {};

    /**
     * Computes the quartet potential for a the given triplet and labels.
     */
    double computeQuartetCost(int quartet, int labelA, int labelB, int labelC, int labelD) { return 0; }

    // for transforming anatomy
    NEWMESH::Triangle deform_anatomy(const int&,const int&,std::map<int,NEWMESH::Pt>  & ,std::map<int,bool> &, std::map<int,NEWMESH::Pt>  &);


    void resample_weights();

    void get_target_data(const int &,const NEWMAT::Matrix &);

    void get_target_data(const int&,const NEWMESH::Pt&,const NEWMESH::Pt&,const NEWMESH::Pt&);



    virtual void get_data_index(const int & controlpoint, const int & sourcepoint, int & ind){}; // data is held in 2D vectors in each case, but for univariate case data is controlpointsxfeatures and for multivariate its sourcepointsxfeatures. This function determines the correct index for rows
  };

  class UnivariateNonLinearSRegDiscreteCostFunction: public NonLinearSRegDiscreteCostFunction
  {
  public:
    //
    // Constructor.
    //
    UnivariateNonLinearSRegDiscreteCostFunction(){};
    virtual void initialize(int numNodes, int numLabels, int numPairs,int numTriplets = 0, int numQuartets = 0); // quartets not used yet
    virtual void get_source_data();

    double computeUnaryCost(int node, int label);

    // void resample_source_data(const int &);
    void reset_target_data(int node){ _targetdata[node-1].clear();}
    virtual void resample_target_data(const NEWMESH::Pt &, const NEWMESH::Pt &, const NEWMESH::Pt &, const NEWMESH::Pt &, const int &, const int &, const int &, int &);
    void get_data_index(const int & controlpoint, const int & sourcepoint, int &ind){ind=controlpoint;}

  };

  class MultivariateNonLinearSRegDiscreteCostFunction: public NonLinearSRegDiscreteCostFunction
  {
  public:
    //
    // Constructor.
    //
    MultivariateNonLinearSRegDiscreteCostFunction(){};
    virtual void initialize(int numNodes,int numLabels, int numPairs, int numTriplets = 0,int numQuartets = 0); // quartets not used yet
    void get_source_data();

    double computeUnaryCost(int node, int label);

    //   void resample_source_data(const int & sourcept);
    void reset_target_data(int node){  for(unsigned int i=0;i<_sourceinrange[node-1].size();i++) { _targetdata[_sourceinrange[node-1][i]-1].clear();}} /// *****NOT THREAD SAFE ********

    virtual void resample_target_data(const NEWMESH::Pt &, const NEWMESH::Pt &, const NEWMESH::Pt &, const NEWMESH::Pt &, const int &, const int &, const int &, int &);
    void get_data_index(const int & controlpoint, const int & sourcepoint, int & ind){if(sourcepoint > (int) _sourcedata.size()){std::cout << "vector index too high " << std::endl;exit(1);}
      ind=sourcepoint;}

  };

  class alphaMINonLinearSRegDiscreteCostFunction: public MultivariateNonLinearSRegDiscreteCostFunction
  {
  public:
    //
    // Constructor.
    //
    alphaMINonLinearSRegDiscreteCostFunction(){};
    //   void resample_source_data(const int & sourcept);

  };

 class HOUnivariateNonLinearSRegDiscreteCostFunction: public UnivariateNonLinearSRegDiscreteCostFunction
  {
  public:
    //
    // Constructor.
    //
    HOUnivariateNonLinearSRegDiscreteCostFunction(){};
    void initialize(int numNodes,int numLabels, int numPairs,int numTriplets = 0, int numQuartets = 0); // quartets not used yet
    void get_source_data();

    double computeUnaryCost(int node, int label){return 0;};

    void set_relations(const std::shared_ptr<NEWMESH::RELATIONS> &CONTROL,const std::shared_ptr<NEWMESH::RELATIONS> &TARG){_controlrel=CONTROL; _targetrel=TARG;
      _sourcerel=_controlrel->invert_relationsTR(_CPgrid,_SOURCE);  }
    double triplet_likelihood(const int &,const int &,const int &,const int &,const NEWMESH::Pt &,const NEWMESH::Pt &,const NEWMESH::Pt &);

  };

 class HOMultivariateNonLinearSRegDiscreteCostFunction: public MultivariateNonLinearSRegDiscreteCostFunction
  {
  public:
    //
    // Constructor.
    //
    HOMultivariateNonLinearSRegDiscreteCostFunction(){};
    void initialize(int numNodes,int numLabels, int numPairs,int numTriplets = 0, int numQuartets = 0); // quartets not used yet
    void get_source_data();

    double computeUnaryCost(int node, int label){return 0;};

    void set_relations(const std::shared_ptr<NEWMESH::RELATIONS> &CONTROL,const std::shared_ptr<NEWMESH::RELATIONS> &TARG){_controlrel=CONTROL; _targetrel=TARG;
      _sourcerel=_controlrel->invert_relationsTR(_CPgrid,_SOURCE); }

    double triplet_likelihood(const int &,const int &,const int &,const int &,const NEWMESH::Pt &,const NEWMESH::Pt &,const NEWMESH::Pt &);
  };



 class RegularisationDiscreteCostFunction: public SRegDiscreteCostFunction
 {
 	 protected:

     //////////////////////// MESHES //////////////////////
    // NEWMESH::newmesh _SPHERE; // SOURCE MESH
     float _rexp;

   public:
     //
     // Constructor.
     //
     RegularisationDiscreteCostFunction(double shear, double bulk, double exp);


	 void initialize(int numNodes,int numLabels, int numPairs,int numTriplets = 0, int numQuartets = 0); // quartets not used yet
	 //void reset_CPgrid(const NEWMESH::newmesh & grid, int num=0){_SPHERE=grid;}
	 double computeTripletRegularisation(int triplet, int labelA, int labelB, int labelC);

 };

 class MetricDistortionDiscreteCostFunction: public RegularisationDiscreteCostFunction
  {

      protected:
      bool isinitialised;

  	  public:
		//
		// Constructor.
		//
		MetricDistortionDiscreteCostFunction(double shear, double bulk, double exp):RegularisationDiscreteCostFunction(shear, bulk,  exp){isinitialised=false;}


		double computeTripletCost(int triplet, int labelA, int labelB, int labelC){
			if (isinitialised==false)
				throw  DISCRETEOPTHOCRException("MetricDistortionDiscreteCostFunction::You must supply meshes.");
			return computeTripletRegularisation(triplet,labelA, labelB,labelC);

		}
};





}
#endif
